
/* idirect3d8.cpp - IDirect3D8 Implementation
 *
 * dxglwrapper 0.08 - Sept. 12, 2002
 * For conditions of distribution and use, see copyright notice in dxgl.h
 * Copyright (c) 1991-2002 realtech VR
 *
 * This file provides implementation of the IDirect3D8 COM Class
 */

#include "dxgl.h"
#include "dxglinit.h"

BIDirect3D8::BIDirect3D8() : m_pGlDevice(0), m_refCount(0)
{
	IAPIDEBUG(Direct3D Pseudo Device);
	GL_CreatePseudoDevice(&m_pGlDevice);
}

ULONG 
BIDirect3D8::AddRef()
{
	MAPIDEBUG(Direct3D::AddRef);
	m_refCount++;
	return m_refCount;
}

HRESULT 
BIDirect3D8::CheckDepthStencilMatch(UINT Adapter, D3DDEVTYPE DeviceType, D3DFORMAT AdapterFormat, D3DFORMAT RenderTargetFormat, D3DFORMAT DepthStencilFormat)
{
	APIDEBUG(CheckDepthStencilMatch);
	return D3D_OK;
}


HRESULT
BIDirect3D8::CheckDeviceFormat(UINT Adapter, D3DDEVTYPE DeviceType, D3DFORMAT AdapterFormat, DWORD Usage, D3DRESOURCETYPE RType, D3DFORMAT CheckFormat)
{	
	APIDEBUG(CheckDeviceFormat);
	// Filter texture formats ( knowns to be compatible )
	// Disable this test for later usage (debug only)
	if (RType  == D3DRTYPE_TEXTURE)
	{
		// supports only true color textures
		if  (  (CheckFormat != D3DFMT_R8G8B8) 
			&& (CheckFormat != D3DFMT_X8R8G8B8)
			&& (CheckFormat != D3DFMT_A8R8G8B8)
			)
			
		return D3DERR_NOTAVAILABLE;
	}

	return D3D_OK;
}

HRESULT
BIDirect3D8::CheckDeviceMultiSampleType(UINT Adapter, D3DDEVTYPE DeviceType, D3DFORMAT SurfaceFormat, BOOL Windowed, D3DMULTISAMPLE_TYPE MultiSampleType)
{
	APIDEBUG(CheckDeviceMultiSampleType);
	return D3D_OK;
}

HRESULT
BIDirect3D8::CheckDeviceType(UINT Adapter, D3DDEVTYPE CheckType, D3DFORMAT DisplayFormat, D3DFORMAT BackBufferFormat, BOOL Windowed)
{
	APIDEBUG(CheckDeviceType);
	return D3D_OK;
}

HRESULT
BIDirect3D8::CreateDevice(UINT Adapter, D3DDEVTYPE DeviceType, HWND hFocusWindow, DWORD BehaviorFlags, D3DPRESENT_PARAMETERS *pPresentationParameters, IDirect3DDevice8 **ppReturnedDeviceInterface)
{
	APIDEBUG(CreateDevice);
	*ppReturnedDeviceInterface = 0;

	if (GL_CreateDevice(&m_pGlDevice, hFocusWindow, Adapter, pPresentationParameters->BackBufferWidth, pPresentationParameters->BackBufferHeight))
		return D3DERR_NOTAVAILABLE;
					
	*ppReturnedDeviceInterface = new BIDirect3DDevice8(this, pPresentationParameters);
	return D3D_OK;
}

HRESULT
BIDirect3D8::EnumAdapterModes(UINT Adapter, UINT Mode, D3DDISPLAYMODE *pMode)
{
	APIDEBUG(EnumAdapterModes);
	memset(pMode, 0, sizeof(D3DDISPLAYMODE));
	UINT bpp;
	GL_EnumAdapterModes(m_pGlDevice, Mode, &pMode->Width, &pMode->Height, &bpp, &pMode->RefreshRate);	

	pMode->Format = D3DFMT_X8R8G8B8;	

	if (bpp == 24)
		pMode->Format = D3DFMT_R8G8B8;
	else
	if (bpp == 16)
		pMode->Format = D3DFMT_R5G6B5;
	else
	if (bpp == 15)
		pMode->Format = D3DFMT_X1R5G5B5;

	return D3D_OK;
}

UINT
BIDirect3D8::GetAdapterCount()
{
	IAPIDEBUG(GetAdapterCount);
	return GL_GetAdapterCount(m_pGlDevice);
}

UINT
BIDirect3D8::GetAdapterDisplayMode(UINT Adapter, D3DDISPLAYMODE *pMode)
{
	IAPIDEBUG(GetAdapterDisplayMode);
	if (!pMode)
		return D3DERR_INVALIDCALL;
	UINT bpp;

	GL_GetAdapterDisplayMode(m_pGlDevice, &pMode->Width, &pMode->Height, &bpp, &pMode->RefreshRate);	

	pMode->Format = D3DFMT_X8R8G8B8;	

	if (bpp == 24)
		pMode->Format = D3DFMT_R8G8B8;
	else
	if (bpp == 16)
		pMode->Format = D3DFMT_R5G6B5;
	else
	if (bpp == 15)
		pMode->Format = D3DFMT_X1R5G5B5;

	return D3D_OK;
}


HRESULT
BIDirect3D8::GetAdapterIdentifier(UINT Adapter, DWORD Flags, D3DADAPTER_IDENTIFIER8 *pIdentifier)
{
	IAPIDEBUG(GetAdapterIdentifier);
	if (!pIdentifier)
		return D3DERR_INVALIDCALL;
		
	memset(pIdentifier, 0, sizeof(D3DADAPTER_IDENTIFIER8));
	GL_GetAdapterName(m_pGlDevice, Adapter, pIdentifier->Driver, MAX_DEVICE_IDENTIFIER_STRING);
	strcpy(pIdentifier->Description,"OpenGL");
#ifdef _WIN32
		pIdentifier->DriverVersion = 0;
#else
		pIdentifier->DriverVersionLowPart = 0;
		pIdentifier->DriverVersionHighPart = 0;
#endif
	pIdentifier->VendorId = 0;
	pIdentifier->DeviceId = 0;
	pIdentifier->SubSysId = 0;
	pIdentifier->Revision = 0;
	pIdentifier->WHQLLevel = 0;	
	return D3D_OK;
}

HMONITOR
BIDirect3D8::GetAdapterMonitor(UINT Adapter)
{
	IAPIDEBUG(GetAdapterMonitor);
	return 0;
}

UINT
BIDirect3D8::GetAdapterModeCount(UINT Adapter)
{
	IAPIDEBUG(GetAdapterModeCount);
	return GL_GetAdapterModeCount(m_pGlDevice);
}

/*
	strncpy(pCaps->szExtensions, (char*)glGetString(GL_EXTENSIONS), GLI_MAX_EXTENSION_SIZE);
	strncpy(pCaps->szRendererName, (char*)glGetString(GL_RENDERER), GLI_MAX_NAME_SIZE);
	strncpy(pCaps->szVersion, (char*)glGetString(GL_VERSION), GLI_MAX_NAME_SIZE);
	strncpy(pCaps->szVendorName, (char*)glGetString(GL_VENDOR), GLI_MAX_NAME_SIZE);

	
	glGetIntegerv(GL_MAX_MODELVIEW_STACK_DEPTH, &pCaps->dwMaxModelViewStackDepth);
	glGetIntegerv(GL_MAX_PROJECTION_STACK_DEPTH, &pCaps->dwMaxProjectionStackDepth);
	glGetIntegerv(GL_MAX_TEXTURE_STACK_DEPTH, &pCaps->dwMaxTextureStackDepth);
	glGetIntegerv(GL_SUBPIXEL_BITS, &pCaps->dwSubPixelBits);
	
	glGetIntegerv(GL_MAX_PIXEL_MAP_TABLE, &pCaps->dwPixelMapTable);
	glGetIntegerv(GL_MAX_NAME_STACK_DEPTH, &pCaps->dwNameStackDepth);
	glGetIntegerv(GL_MAX_LIST_NESTING, &pCaps->dwListNesting);
	glGetIntegerv(GL_MAX_EVAL_ORDER, &pCaps->dwEvalOrder);
	glGetIntegerv(GL_MAX_VIEWPORT_DIMS, &pCaps->dwViewportMaxWidth);
	glGetIntegerv(GL_MAX_ATTRIB_STACK_DEPTH, &pCaps->dwAttribStackDepth);
	
	glGetFloatv(GL_LINE_WIDTH_RANGE, &pCaps->fLineSizeMin);
	glGetFloatv(GL_LINE_WIDTH_GRANULARITY, &pCaps->fLineSizeGran);

	glGetBooleanv(GL_AUX_BUFFERS, &pCaps->bAuxBuffers);
	glGetBooleanv(GL_RGBA_MODE, &pCaps->bRGBAMode);
	glGetBooleanv(GL_INDEX_MODE, &pCaps->bIndexMode);
	glGetBooleanv(GL_DOUBLEBUFFER, &pCaps->bDoubleBuffer);
	glGetBooleanv(GL_STEREO, &pCaps->bStereo);

	*/


// in 0.04 : GetDeviceCaps empty the structure and fooled d3dx8 ?
HRESULT
BIDirect3D8::GetDeviceCaps(UINT Adapter, D3DDEVTYPE DeviceType, D3DCAPS8 *pCaps)
{
	IAPIDEBUG(GetDeviceCaps);

	GLint n;
	GLfloat r[2];	
	if (!pCaps)
		return D3DERR_INVALIDCALL;
	memset(pCaps, 0, sizeof(D3DCAPS8));

	   /* Device Info */
    pCaps->DeviceType = D3DDEVTYPE_HAL;
    pCaps->AdapterOrdinal = 0;

    /* Caps from DX7 Draw */
    pCaps->Caps = 0; //D3DCAPS_READ_SCANLINE;
    pCaps->Caps2 = D3DCAPS2_CANRENDERWINDOWED ;
    pCaps->Caps3 = 0;
    pCaps->PresentationIntervals = D3DPRESENT_INTERVAL_IMMEDIATE; // see wglSwapIntervalEXT

    /* Cursor Caps */
    pCaps->CursorCaps = 0;

    /* 3D Device Caps */
    pCaps->DevCaps = D3DDEVCAPS_EXECUTESYSTEMMEMORY
					|D3DDEVCAPS_EXECUTEVIDEOMEMORY
					|D3DDEVCAPS_TLVERTEXSYSTEMMEMORY
					|D3DDEVCAPS_DRAWPRIMTLVERTEX
					|D3DDEVCAPS_CANRENDERAFTERFLIP
					|D3DDEVCAPS_TEXTURENONLOCALVIDMEM
					|D3DDEVCAPS_DRAWPRIMITIVES2
					|D3DDEVCAPS_DRAWPRIMITIVES2EX
					|D3DDEVCAPS_HWTRANSFORMANDLIGHT
					|D3DDEVCAPS_HWRASTERIZATION
					|D3DDEVCAPS_PUREDEVICE					
					;

    pCaps->PrimitiveMiscCaps =
					D3DPMISCCAPS_MASKZ|
					D3DPMISCCAPS_CULLNONE|
					D3DPMISCCAPS_CULLCW|
					D3DPMISCCAPS_CULLCCW|
					D3DPMISCCAPS_COLORWRITEENABLE|
					D3DPMISCCAPS_TSSARGTEMP|
					D3DPMISCCAPS_BLENDOP					
					;

    pCaps->RasterCaps = 
					D3DPRASTERCAPS_DITHER|
					D3DPRASTERCAPS_FOGVERTEX|
					D3DPRASTERCAPS_FOGTABLE|
					D3DPRASTERCAPS_MIPMAPLODBIAS|
					D3DPRASTERCAPS_ZBIAS|
					D3DPRASTERCAPS_FOGRANGE|
					D3DPRASTERCAPS_ANISOTROPY|	// Check extensions
					D3DPRASTERCAPS_WFOG|		
					D3DPRASTERCAPS_ZFOG					
					;




    pCaps->ZCmpCaps = 
					D3DPCMPCAPS_NEVER|
					D3DPCMPCAPS_LESS|
					D3DPCMPCAPS_EQUAL|
					D3DPCMPCAPS_LESSEQUAL|
					D3DPCMPCAPS_GREATER|
					D3DPCMPCAPS_NOTEQUAL|
					D3DPCMPCAPS_GREATEREQUAL|
					D3DPCMPCAPS_ALWAYS
					;


    pCaps->SrcBlendCaps = 
					D3DPBLENDCAPS_ZERO|
					D3DPBLENDCAPS_ONE|
					D3DPBLENDCAPS_SRCCOLOR|
					D3DPBLENDCAPS_INVSRCCOLOR|
					D3DPBLENDCAPS_SRCALPHA|
					D3DPBLENDCAPS_INVSRCALPHA|
					D3DPBLENDCAPS_INVDESTALPHA|
					D3DPBLENDCAPS_DESTCOLOR|
					D3DPBLENDCAPS_INVDESTCOLOR|
					D3DPBLENDCAPS_SRCALPHASAT|
					D3DPBLENDCAPS_BOTHSRCALPHA|
					D3DPBLENDCAPS_BOTHINVSRCALPHA
					;

    pCaps->DestBlendCaps = 
					D3DPBLENDCAPS_ZERO|
					D3DPBLENDCAPS_ONE|
					D3DPBLENDCAPS_SRCCOLOR|
					D3DPBLENDCAPS_INVSRCCOLOR|
					D3DPBLENDCAPS_SRCALPHA|
					D3DPBLENDCAPS_INVSRCALPHA|
					D3DPBLENDCAPS_INVDESTALPHA|
					D3DPBLENDCAPS_DESTCOLOR|
					D3DPBLENDCAPS_INVDESTCOLOR|
					D3DPBLENDCAPS_SRCALPHASAT|
					D3DPBLENDCAPS_BOTHSRCALPHA|
					D3DPBLENDCAPS_BOTHINVSRCALPHA
					;
    pCaps->AlphaCmpCaps = 
					D3DPCMPCAPS_NEVER|
					D3DPCMPCAPS_LESS|
					D3DPCMPCAPS_EQUAL|
					D3DPCMPCAPS_LESSEQUAL|
					D3DPCMPCAPS_GREATER|
					D3DPCMPCAPS_NOTEQUAL|
					D3DPCMPCAPS_GREATEREQUAL|
					D3DPCMPCAPS_ALWAYS
					;

    pCaps->ShadeCaps = 
					D3DPSHADECAPS_COLORGOURAUDRGB|
					D3DPSHADECAPS_SPECULARGOURAUDRGB|
					D3DPSHADECAPS_ALPHAGOURAUDBLEND|
					D3DPSHADECAPS_FOGGOURAUD
					;

    pCaps->TextureCaps = 
					D3DPTEXTURECAPS_PERSPECTIVE|
					D3DPTEXTURECAPS_POW2|
					D3DPTEXTURECAPS_ALPHA|
					D3DPTEXTURECAPS_ALPHAPALETTE|
					D3DPTEXTURECAPS_PROJECTED|
					D3DPTEXTURECAPS_CUBEMAP|		// Check extension
					D3DPTEXTURECAPS_MIPMAP|		// Check extension
					D3DPTEXTURECAPS_MIPCUBEMAP|	// Check extension
					D3DPTEXTURECAPS_CUBEMAP_POW2// Check extension					
					;		
    pCaps->TextureFilterCaps =           // D3DPTFILTERCAPS for IDirect3DTexture8's
					D3DPTFILTERCAPS_MINFPOINT|
					D3DPTFILTERCAPS_MINFLINEAR|
					D3DPTFILTERCAPS_MINFANISOTROPIC|	// Check
					D3DPTFILTERCAPS_MIPFPOINT|
					D3DPTFILTERCAPS_MIPFLINEAR|
					D3DPTFILTERCAPS_MAGFPOINT|
					D3DPTFILTERCAPS_MAGFLINEAR
					;

		
    pCaps->CubeTextureFilterCaps = 0;      // D3DPTFILTERCAPS for IDirect3DCubeTexture8's
    pCaps->VolumeTextureFilterCaps = 0;    // D3DPTFILTERCAPS for IDirect3DVolumeTexture8's
    pCaps->TextureAddressCaps =		       // D3DPTADDRESSCAPS for IDirect3DTexture8's
					D3DPTADDRESSCAPS_WRAP|
					D3DPTADDRESSCAPS_MIRROR|	// check extension
					D3DPTADDRESSCAPS_CLAMP|	// check extension
					D3DPTADDRESSCAPS_INDEPENDENTUV	// check extension
					;
			
    pCaps->VolumeTextureAddressCaps = 0;   // D3DPTADDRESSCAPS for IDirect3DVolumeTexture8's

    pCaps->LineCaps =                    // D3DLINECAPS
					D3DLINECAPS_TEXTURE|
					D3DLINECAPS_ZTEST|
					D3DLINECAPS_BLEND|
					D3DLINECAPS_ALPHACMP|
					D3DLINECAPS_FOG
					;

	
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n);

    pCaps->MaxTextureWidth = n; 
	pCaps->MaxTextureHeight = n;			// would explain d3dx retained mode crash ?
    pCaps->MaxVolumeExtent = 0;

    pCaps->MaxTextureRepeat = 8192;			// Geforce 2MX
    pCaps->MaxTextureAspectRatio = 0;		// ?

#if GL_EXT_texture_filter_anisotropic
	glGetIntegerv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &n);
    pCaps->MaxAnisotropy = n;
#else
	pCaps->MaxAnisotropy = 0;
#endif

    pCaps->MaxVertexW = 1E+010;

    pCaps->GuardBandLeft = -1E+008;
    pCaps->GuardBandTop = -1E+008;
    pCaps->GuardBandRight = 1E+008;
    pCaps->GuardBandBottom = 1E+008;

    pCaps->ExtentsAdjust = 0;
    pCaps->StencilCaps = 
					D3DSTENCILCAPS_KEEP|
					D3DSTENCILCAPS_ZERO|
					D3DSTENCILCAPS_REPLACE|
					D3DSTENCILCAPS_INCRSAT|	// check extension
					D3DSTENCILCAPS_DECRSAT|	// check extension
					D3DSTENCILCAPS_INVERT|	// check extension
					D3DSTENCILCAPS_INCR|	// check extension
					D3DSTENCILCAPS_DECR	// check extension
					;
    pCaps->FVFCaps = D3DFVFCAPS_DONOTSTRIPELEMENTS;
    pCaps->TextureOpCaps = 
					D3DTEXOPCAPS_DISABLE|
					D3DTEXOPCAPS_SELECTARG1|	// check extension
					D3DTEXOPCAPS_SELECTARG2|	// check extension
					D3DTEXOPCAPS_MODULATE|		// check extension
					D3DTEXOPCAPS_MODULATE2X|	// check extension
					D3DTEXOPCAPS_MODULATE4X|	// check extension
					D3DTEXOPCAPS_ADD|			// check extension
					D3DTEXOPCAPS_ADDSIGNED|	// check extension
					D3DTEXOPCAPS_ADDSIGNED2X|	// check extension
					D3DTEXOPCAPS_SUBTRACT|	// check extension
					D3DTEXOPCAPS_ADDSMOOTH|	// check extension
					D3DTEXOPCAPS_ADDSIGNED|	// check extension
					D3DTEXOPCAPS_BLENDDIFFUSEALPHA|	// check extension
					D3DTEXOPCAPS_BLENDTEXTUREALPHA|	// check extension
					D3DTEXOPCAPS_BLENDFACTORALPHA|	// check extension
					D3DTEXOPCAPS_BLENDTEXTUREALPHAPM|	// check extension
					D3DTEXOPCAPS_BLENDCURRENTALPHA|	// check extension
					D3DTEXOPCAPS_MODULATEALPHA_ADDCOLOR|	// check extension
					D3DTEXOPCAPS_MODULATECOLOR_ADDALPHA|	// check extension
					D3DTEXOPCAPS_MODULATEINVALPHA_ADDCOLOR|	// check extension
					D3DTEXOPCAPS_MODULATEINVCOLOR_ADDALPHA|	// check extension
					D3DTEXOPCAPS_DOTPRODUCT3|	// check extension
					D3DTEXOPCAPS_MULTIPLYADD|	// check extension
					D3DTEXOPCAPS_LERP	// check extension
					;


    pCaps->MaxTextureBlendStages = 8;

#if GL_ARB_multitexture
	glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &n);
    pCaps->MaxSimultaneousTextures = n;
#else
	pCaps->MaxSimultaneousTextures = 1;
#endif

    pCaps->VertexProcessingCaps = 
					D3DVTXPCAPS_DIRECTIONALLIGHTS|
					D3DVTXPCAPS_LOCALVIEWER|
					D3DVTXPCAPS_POSITIONALLIGHTS|
					D3DVTXPCAPS_TEXGEN
					;	
	
	glGetIntegerv(GL_MAX_LIGHTS, &n);
    pCaps->MaxActiveLights = n;

	glGetIntegerv(GL_MAX_CLIP_PLANES, &n);
    pCaps->MaxUserClipPlanes = n;

	n = 0;
#if GL_ARB_vertex_blend
	glGetIntegerv(GL_MAX_VERTEX_UNITS_ARB, &n);	
#endif

	pCaps->MaxVertexBlendMatrices = n;		//	GL_MAX_VERTEX_UNITS_ARB ?
    pCaps->MaxVertexBlendMatrixIndex = 0;
	
	glGetFloatv(GL_POINT_SIZE_RANGE, r);
    pCaps->MaxPointSize = r[1];

    pCaps->MaxPrimitiveCount = 65535;       // max number of primitives per DrawPrimitive call
    pCaps->MaxVertexIndex = 65535;			// 16bit max for now
    pCaps->MaxStreams = 1 ;					// Direct3D 7 like
    pCaps->MaxStreamStride = 65535;			// max stride for SetStreamSource

    pCaps->VertexShaderVersion = 0x100;
    pCaps->MaxVertexShaderConst = 0;		// number of vertex shader constant registers

    pCaps->PixelShaderVersion = 0x100;
    pCaps->MaxPixelShaderValue = 0;        // max value of pixel shader arithmetic component


	return D3D_OK;
}

HRESULT
BIDirect3D8::QueryInterface(REFIID riid, void **ppvObj)
{
	IAPIDEBUG(QueryInterface);
	// COM Interface. Anyway, real Direct3D doesnt support	
	*ppvObj = 0;
	return D3DERR_INVALIDCALL;
}

HRESULT
BIDirect3D8::RegisterSoftwareDevice(void *pInitializeFunction)
{
	APIDEBUG(RegisterSoftwareDevice);
	return D3D_OK;
}

ULONG
BIDirect3D8::Release()
{		
	MAPIDEBUG(Direct3D::Release);
	m_refCount--;
	if ( m_refCount < 0 )	
	{
		MAPIDEBUG(Direct3D::DELETE);
		GL_ReleaseDevice(m_pGlDevice);
		m_pGlDevice = 0;
		delete this;
	}	
	return D3D_OK;
}
