
/* idirect3dtexture8.cpp - IDirect3DTexture8
 *
 * dxglwrapper 0.08 - Sept. 12, 2002
 * For conditions of distribution and use, see copyright notice in dxgl.h
 * Copyright (c) 1991-2002 realtech VR
 *
 * This file provides implementation of the IDirect3DTexture8 COM Class
 */
#include "dxgl.h"
#include "win32/wglext.h"

GLenum D3D8ToGLEnum(D3DFORMAT fmt)
{
	return GL_RGBA;
}

class BIDirect3DSurface8_itex : public BIDirect3DSurface8
{
public:
	BIDirect3DSurface8_itex(BIDirect3DTexture8 *pTex) 
		: BIDirect3DSurface8(pTex->m_pDevice, pTex->m_Width, pTex->m_Height, pTex->m_Format, D3DMULTISAMPLE_NONE, TRUE), m_pTex(pTex)
	{
		pTex->AddRef();
}	

	STDMETHOD(LockRect)(D3DLOCKED_RECT *pLockedRect, CONST RECT *pRect, DWORD Flags)
	{
		return m_pTex->LockRect(0, pLockedRect, pRect, Flags);
}

	STDMETHOD_(ULONG, Release)()
	{
		return m_pTex->Release();
}

	STDMETHOD(UnlockRect)()
	{
		return m_pTex->UnlockRect(0);
}

	STDMETHOD(GetContainer)(REFIID iid, void **ppContainer)
	{
		APIDEBUG(Surface_Itex::GetContainer);
		m_pTex->AddRef();
		* ppContainer = m_pTex;
		return D3D_OK;
}

	BIDirect3DTexture8	*	m_pTex;
};

struct PixelTag 
{
	enum _D3DFORMAT	Format;
	int				dwSize;
	GLenum 			internalFmt;
	GLenum 			textureFmt;
};

static struct PixelTag DX_Fmt[] = { 
	{ D3DFMT_R8G8B8 		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_A8R8G8B8		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_X8R8G8B8		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_A8 			, 1 , GL_ALPHA8 , GL_ALPHA}, 
	{ D3DFMT_A8R3G3B2		, 2 , GL_RGBA4 , GL_BGRA_EXT}, 
	{ D3DFMT_X4R4G4B4		, 2 , GL_RGBA4 , GL_BGRA_EXT}, 
	{ D3DFMT_A2B10G10R10	, 4 , GL_RGB10_A2 , GL_BGRA_EXT}, 
	{ D3DFMT_G16R16 		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 

	{ D3DFMT_R5G6B5 		, 2 , GL_UNSIGNED_SHORT_5_6_5_EXT , GL_BGR_EXT}, 
	{ D3DFMT_X1R5G5B5		, 2 , GL_UNSIGNED_SHORT_5_5_5_1_EXT, GL_BGR_EXT}, 
	{ D3DFMT_A1R5G5B5		, 2 , GL_UNSIGNED_SHORT_5_5_5_1_EXT, GL_BGR_EXT}, 
	{ D3DFMT_A4R4G4B4		, 2 , GL_UNSIGNED_SHORT_4_4_4_4_EXT, GL_BGRA_EXT}, 
	{ D3DFMT_R3G3B2 		, 1 , GL_UNSIGNED_BYTE_3_3_2_EXT, GL_BGR_EXT}, 

	{ D3DFMT_A8P8			, 2 , GL_LUMINANCE8_ALPHA8 , GL_LUMINANCE_ALPHA}, 
	{ D3DFMT_P8 			, 1 , GL_COLOR_INDEX , GL_COLOR_INDEX}, 

	{ D3DFMT_L8 			, 1 , GL_LUMINANCE8 , GL_LUMINANCE}, 
	{ D3DFMT_A8L8			, 2 , GL_LUMINANCE8_ALPHA8 , GL_LUMINANCE}, 
	{ D3DFMT_A4L4			, 1 , GL_LUMINANCE4_ALPHA4 , GL_LUMINANCE}, 

	{ D3DFMT_V8U8			, 2 , GL_RGBA4 , GL_BGRA_EXT}, 		// GL_DU8DV8_ATI
	{ D3DFMT_L6V5U5 		, 2 , GL_RGBA4 , GL_BGRA_EXT}, 	
	{ D3DFMT_X8L8V8U8		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_Q8W8V8U8		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_V16U16 		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_W11V11U10		, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_A2W10V10U10	, 4 , GL_RGBA8 , GL_BGRA_EXT}, 

	{ D3DFMT_UYVY			, 4 , GL_RGBA8 , GL_BGRA_EXT}, 
	{ D3DFMT_YUY2			, 4 , GL_RGBA8 , GL_BGRA_EXT}, 

	// WARNING : Some extensions are required (EXT_packed_pixels and ARB_Texture_Compression
	{ D3DFMT_DXT1			, 4 , GL_COMPRESSED_RGB_S3TC_DXT1_EXT , GL_BGRA_EXT}, 
	{ D3DFMT_DXT2			, 4 , GL_COMPRESSED_RGBA_S3TC_DXT1_EXT , GL_BGRA_EXT}, 
	{ D3DFMT_DXT3			, 4 , GL_COMPRESSED_RGBA_S3TC_DXT3_EXT , GL_BGRA_EXT}, 
	{ D3DFMT_DXT4			, 4 , GL_COMPRESSED_RGBA_S3TC_DXT5_EXT , GL_BGRA_EXT}, 
	{ D3DFMT_DXT5			, 4 , GL_COMPRESSED_RGBA_S3TC_DXT5_EXT , GL_BGRA_EXT}, 

	{ D3DFMT_UNKNOWN		, 0 , 0 , 0}

};


BIDirect3DTexture8::BIDirect3DTexture8(IDirect3DDevice8 *pDevice, UINT Width, UINT Height, UINT Levels, DWORD Usage, D3DFORMAT Format, D3DPOOL Pool) :
m_pDevice(pDevice), m_Width(Width), m_Height(Height), m_Levels(Levels), m_Usage(Usage), m_Format(Format), m_Pool(Pool), m_refCount(0), 
m_InternalFmt(0), m_TextureFmt(0), m_pLocalImage(0), m_Pitch(0)
{	
	MAPIDEBUG(new Texture);

	GL_DEBUG_TRY();	

	struct PixelTag * p = DX_Fmt;
	m_Pitch = 4 * m_Width;
	while(p->Format!=D3DFMT_UNKNOWN)
	{
		if (p->Format == Format)
		{
			m_InternalFmt = p->internalFmt;
			m_TextureFmt = p->textureFmt;
			m_Pitch = p->dwSize * m_Width;
			
			break;
	}
		p++;
	}
	
	m_pLocalImage = new char[m_Width * m_Pitch];
	ASSERT(m_pLocalImage);
	memset(m_pLocalImage, 0xcd, m_Width * m_Pitch);

	m_GlHandle = 0;

	if (dxgl_option & DXGL_OPTION_NOTEXTURE)
		return;

	glGenTextures(1, &m_GlHandle);	
	glBindTexture(GL_TEXTURE_2D, m_GlHandle);

	GL_DEBUG_CATCH();
}

	

void
BIDirect3DTexture8::BindTexture()
{
	RAPIDEBUG(BindTexture);

	if (dxgl_option & DXGL_OPTION_NOTEXTURE)
		return;

	GL_DEBUG_TRY();

	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, m_GlHandle);

	GL_DEBUG_CATCH();
}

ULONG
BIDirect3DTexture8::AddRef()
{
	MAPIDEBUG(Texture::AddRef);
	m_refCount++;
	return 0;
}

HRESULT 
BIDirect3DTexture8::FreePrivateData(REFGUID refguid)
{
	APIDEBUG(FreePrivateData);
	return D3D_OK;
}

HRESULT
BIDirect3DTexture8::GetDevice(IDirect3DDevice8 **ppDevice)
{
	IAPIDEBUG(GetDevice);
	m_pDevice->AddRef();
	*ppDevice = m_pDevice;	
	return D3D_OK;
}

DWORD
BIDirect3DTexture8::GetLevelCount()
{
	IAPIDEBUG(GetLevelCount);
	return m_Levels;
}

DWORD
BIDirect3DTexture8::GetLOD()
{
	IAPIDEBUG(GetLOD);
	return m_Levels;
}

DWORD
BIDirect3DTexture8::GetPriority()
{
	IAPIDEBUG(GetPriority);
	return m_Priority;
}

HRESULT
BIDirect3DTexture8::GetPrivateData(REFGUID refguid, void *pData, DWORD *pSizeOfData)
{
	// TODO: Handle this com interface
	IAPIDEBUG(GetPrivateData);
	*pSizeOfData = 0;
	return D3D_OK;
}

D3DRESOURCETYPE
BIDirect3DTexture8::GetType()
{
	IAPIDEBUG(GetType);
	return D3DRTYPE_TEXTURE;
}

void
BIDirect3DTexture8::PreLoad()
{
	APIDEBUG(PreLoad);
	return ;
}

HRESULT
BIDirect3DTexture8::QueryInterface(REFIID, void **ppvObj)
{
	// TODO: Handle this com interface - should always return an error
	APIDEBUG(QueryInterface);
	*ppvObj = 0;
	return D3D_OK;
}

ULONG
BIDirect3DTexture8::Release()
{
	MAPIDEBUG(Texture::Release);
#ifndef __BEOS__
	__try 
#endif
	{
		GL_DEBUG_TRY();
		m_refCount--;

		if ( m_refCount < 0)
		{
			MAPIDEBUG(Texture::DELETE); 
			if (m_GlHandle)
			{
				glDeleteTextures(1, &m_GlHandle);
				m_GlHandle = 0;				
			}

			if (m_pLocalImage)
			{
				delete[] m_pLocalImage;
				m_pLocalImage = 0;
			}
		}	
		GL_DEBUG_CATCH();

	}
#ifndef __BEOS__
	__except(1, 1){
		APIDEBUG(EXCEPTION CATCHED);
	}
#endif
	return D3D_OK;
}

DWORD
BIDirect3DTexture8::SetPriority(DWORD PriorityNew)
{
	APIDEBUG(SetPriority);
	m_Priority = PriorityNew;	
	float v = (float)PriorityNew / 255.f;
	GL_DEBUG_TRY();
	glPrioritizeTextures(1, &m_GlHandle, &v);
	GL_DEBUG_CATCH();
	return D3D_OK;
}

HRESULT
BIDirect3DTexture8::SetPrivateData(REFGUID refguid, CONST void *pData, DWORD SizeOfData, DWORD Flags)
{
	APIDEBUG(SetPrivateData);
	return D3D_OK;
}

HRESULT BIDirect3DTexture8::GetLevelDesc(UINT Level, D3DSURFACE_DESC *pDesc)
{	
	IAPIDEBUG(GetLevelDesc);
	if (Level>m_Levels)
		return D3DERR_INVALIDCALL;

	memset(pDesc, 0, sizeof(pDesc));
	pDesc->Format = m_Format;
	pDesc->Height = m_Height>>Level;
	pDesc->Width = m_Width>>Level;

	return D3D_OK;
}

HRESULT BIDirect3DTexture8::LockRect(UINT Level, D3DLOCKED_RECT *pLockedRect, CONST RECT *pRect, DWORD Flags)
{
	APIDEBUG(Texture::LockRect);
	
	ASSERT(m_Pitch);
	pLockedRect->pBits = m_pLocalImage;
	pLockedRect->Pitch = m_Pitch;
	
	return D3D_OK;
}

HRESULT BIDirect3DTexture8::GetSurfaceLevel(UINT Level, IDirect3DSurface8 **ppSurfaceLevel)
{
	IAPIDEBUG(GetSurfaceLevel);
	if ((Level>m_Levels)||(!ppSurfaceLevel))
		return D3DERR_INVALIDCALL;
	if( Level == 0 )
		*ppSurfaceLevel = new BIDirect3DSurface8_itex(this);
	else
		*ppSurfaceLevel = new BIDirect3DSurface8(m_pDevice, m_Width, m_Height, m_Format, D3DMULTISAMPLE_NONE, TRUE);
	return D3D_OK; // not supported
}

static void *Convert16To24(int width, int height, void *src, int r, int g, int b)
{
	int n = width * height;
	unsigned short *s = (unsigned short *)src;
	unsigned char  *d = (unsigned char* )malloc(n * 3);
	void *_d = d;
	int mr = (1<<r)-1;
	int mg = (1<<g)-1;
	int mb = (1<<b)-1;
	int sr = 8 - mr;
	int sg = 8 - mg;
	int sb = 8 - mb;

	while(n)
	{
		unsigned S = *s;
		d[0] = (S & mr)<<sr; S>>=r;
		d[1] = (S & mg)<<sg; S>>=g;
		d[2] = (S & mb)<<sb; S>>=b;
		n--;
		d+=3;
		s++;
	}
	return _d;
}

static void *Convert16To32(int width, int height, void *src, int r, int g, int b, int a)
{

	int n = width * height;
	unsigned short *s = (unsigned short *)src;
	unsigned char  *d = (unsigned char *) malloc(n * 4);
	void *_d = d;
	int mr = (1<<r)-1;
	int mg = (1<<g)-1;
	int mb = (1<<b)-1;
	int ma = (1<<a)-1;
	int sr = 8 - mr;
	int sg = 8 - mg;
	int sb = 8 - mb;
	int sa = 8 - ma;
	while(n)
	{
		unsigned S = *s;
		d[0] = (S & mr)<<sr; S>>=r;
		d[1] = (S & mg)<<sg; S>>=g;
		d[2] = (S & mb)<<sb; S>>=b;
		d[3] = (S & ma)<<sa;
		n--;
		d+=3;
		s++;
	}
	return _d;
}
		
		

HRESULT BIDirect3DTexture8::UnlockRect(UINT Level)
{
	APIDEBUG(Texture::UnlockRect);
	GL_DEBUG_TRY();

	if (dxgl_option & DXGL_OPTION_NOTEXTURE)
		return D3D_OK;

	glEnable( GL_TEXTURE_2D);
	glBindTexture( GL_TEXTURE_2D, m_GlHandle );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	glPixelStorei( GL_PACK_ROW_LENGTH, 0);
	glPixelStorei( GL_PACK_SKIP_PIXELS, 0); 	

	if ((m_InternalFmt == GL_UNSIGNED_SHORT_5_6_5_EXT)
		||(m_InternalFmt == GL_UNSIGNED_SHORT_5_5_5_1_EXT)
		  ||(m_InternalFmt == GL_UNSIGNED_SHORT_4_4_4_4_EXT))
	{
		// OpenGL 1.2 flaw. there is no 16bit.
		void *tmp;
		int b = 0;
		switch(m_InternalFmt)
		{
			case GL_UNSIGNED_SHORT_5_6_5_EXT:
				tmp = Convert16To24(m_Width, m_Height, m_pLocalImage, 5, 6, 5);
				b = GL_RGB;
				break;
			case GL_UNSIGNED_SHORT_5_5_5_1_EXT:
				tmp = Convert16To32(m_Width, m_Height, m_pLocalImage, 5, 5, 5, 1);
				b = GL_RGBA;
				break;
			case GL_UNSIGNED_SHORT_4_4_4_4_EXT:
				tmp = Convert16To32(m_Width, m_Height, m_pLocalImage, 4, 4, 4, 4);
				b = GL_RGBA;
				break;
		}		

		glTexImage2D( GL_TEXTURE_2D, 
			Level, 						// base level
			GL_RGB,
			m_Width, 
			m_Height, 
			0, 							// border
			m_TextureFmt, 				// texture format
			GL_UNSIGNED_BYTE, 			// texture type
			tmp
		);
		free(tmp);
	}
	else
	glTexImage2D( GL_TEXTURE_2D, 
		Level, 						// base level
		m_InternalFmt, 				// number of color components in the texture
		m_Width, 
		m_Height, 
		0, 							// border
		m_TextureFmt, 				// texture format
		GL_UNSIGNED_BYTE, 			// texture type
		m_pLocalImage
	);

	GL_DEBUG_CATCH();

	
	return D3D_OK;
}

HRESULT BIDirect3DTexture8::AddDirtyRect(CONST RECT *pDirtyRect)
{
	APIDEBUG(AddDirtyRect);
	return D3D_OK;
}

DWORD BIDirect3DTexture8::SetLOD(DWORD lod)
{	
	APIDEBUG(SetLOD);
	m_Lod = lod;
	return m_Lod;
}
